import sys

import SimpleITK as sitk
import numpy as np

from ..image import Image
from ..toolbox_logic import handle_uncaught_exception

sys.excepthook = handle_uncaught_exception


class Preprocessing:

    def __init__(self, input_imaging_modality,
                 resample_resolution=None, resample_dimension=None,
                 interpolation_method=None, interpolation_threshold=None):
        self.input_imaging_modality = input_imaging_modality

        if not (isinstance(resample_resolution, (float, int)) and resample_resolution > 0):
            raise ValueError(f'Resample resolution {resample_resolution} must be a positive int or float.')
        if resample_dimension not in ['3D', '2D']:
            raise ValueError(f"Resample dimension '{resample_dimension}' is not '2D' or '3D'.")

        self.resample_resolution = resample_resolution
        self.resample_dimension = resample_dimension
        self.interpolation_method = interpolation_method
        self.interpolation_threshold = interpolation_threshold

    @staticmethod
    def get_interpolator(interpolation_method):
        """Returns the appropriate SimpleITK interpolator based on the method name."""
        interpolator_mapping = {
            'Linear': sitk.sitkLinear,
            'NN': sitk.sitkNearestNeighbor,
            'BSpline': sitk.sitkBSpline,
            'Gaussian': sitk.sitkGaussian
        }
        if interpolation_method not in interpolator_mapping:
            raise ValueError(f"Interpolation method '{interpolation_method}' is not supported.")
        return interpolator_mapping[interpolation_method]

    @staticmethod
    def calculate_resampled_origin(initial_shape, initial_spacing, resulted_spacing, initial_origin, axis=0):
        """Calculates the resampled image origin for a given axis based on spacing and shape."""
        n_a = float(initial_shape[axis])
        s_a = initial_spacing[axis]
        s_b = resulted_spacing[axis]
        n_b = np.ceil((n_a * s_a) / s_b)
        x_b = initial_origin[axis] + (s_a * (n_a - 1) - s_b * (n_b - 1)) / 2
        return x_b

    def resample(self, image, image_type):
        """Resamples the given image or mask according to the specified resolution and interpolation method."""

        # Calculate output spacing
        if self.resample_dimension == '3D':
            output_spacing = [self.resample_resolution] * 3
        elif self.resample_dimension == '2D':  # '2D' case
            output_spacing = [self.resample_resolution] * 2 + [image.spacing[2]]
        else:
            raise ValueError(f"Resample dimension '{self.resample_dimension}' is not supported.")

        # Calculate output origin
        output_origin = [
            self.calculate_resampled_origin(image.shape, image.spacing, output_spacing, image.origin, axis)
            for axis in range(3)
        ]

        # Calculate output shape
        output_shape = np.ceil((np.array(image.shape) * (np.array(image.spacing) / np.array(output_spacing)))).astype(
            int)

        # Set up SimpleITK image
        sitk_image = sitk.GetImageFromArray(image.array)
        sitk_image.SetSpacing(image.spacing)
        sitk_image.SetOrigin(image.origin)
        sitk_image.SetDirection(image.direction)

        # Configure resampling filter
        resample_filter = sitk.ResampleImageFilter()
        resample_filter.SetOutputSpacing(output_spacing)
        resample_filter.SetOutputOrigin(output_origin)
        resample_filter.SetOutputDirection(image.direction)
        resample_filter.SetSize(output_shape.tolist())
        resample_filter.SetOutputPixelType(sitk.sitkFloat64)
        resample_filter.SetInterpolator(self.get_interpolator(self.interpolation_method))

        # Execute resampling
        resampled_sitk_image = resample_filter.Execute(sitk_image)

        # Process resampled image or mask
        if image_type == 'image':
            resampled_sitk_image = self.process_resampled_image(resampled_sitk_image)
        elif image_type == 'mask':
            resampled_sitk_image = self.process_resampled_mask(resampled_sitk_image)

        resampled_array = sitk.GetArrayFromImage(resampled_sitk_image)

        # Create a new Image instance with resampled data
        return Image(array=resampled_array, origin=output_origin, spacing=output_spacing,
                     direction=image.direction, shape=output_shape)

    def process_resampled_image(self, resampled_sitk_image):
        """Processes resampled image according to modality-specific requirements."""
        if self.input_imaging_modality == 'CT':
            resampled_sitk_image = sitk.Round(resampled_sitk_image)
            resampled_sitk_image = sitk.Cast(resampled_sitk_image, sitk.sitkInt16)
        elif self.input_imaging_modality in ['MR', 'PT']:
            resampled_sitk_image = sitk.Cast(resampled_sitk_image, sitk.sitkFloat64)
        return resampled_sitk_image

    def process_resampled_mask(self, resampled_sitk_image):
        """Processes resampled mask using thresholding."""
        mask_array = sitk.GetArrayFromImage(resampled_sitk_image)
        mask_array = np.where(mask_array >= self.interpolation_threshold, 1, 0).astype(np.int16)
        return sitk.GetImageFromArray(mask_array)



import csv
from pathlib import Path

import numpy as np
import pytest

from zrad.filtering import Filtering
from zrad.image import Image
from zrad.preprocessing import Preprocessing
from zrad.radiomics import Radiomics


def _run_ph_i_case(filtering, phantom, filename, config, data_dir):
    filtered_image = filtering.apply_filter(phantom)
    response_map = Image()
    response_map_path = data_dir / 'Ph_I' / 'response_maps' / filename
    response_map.read_nifti_image(str(response_map_path))
    ibsi_ii_ph_i_validation(filtered_image.array, response_map.array, config)


def ibsi_ii_feature_tolerances(filter_id):
    csv_path = Path(__file__).parent / 'data' / 'ibsi_2_reference_values.csv'
    with open(csv_path, newline='') as csv_file:
        reader = csv.DictReader(csv_file, delimiter=';')
        return {row['feature_tag']: row for row in reader if row['filter_id'] == filter_id}


def ibsi_ii_ph_i_validation(filtered_image, response_map, config_id):

    tolerance = 0.01 * (np.max(response_map) - np.min(response_map))
    within_tolerance = (filtered_image >= (response_map - tolerance)) & (
                filtered_image <= (response_map + tolerance))

    total_voxels = response_map.size
    voxels_within_tolerance = np.sum(within_tolerance)
    if total_voxels != voxels_within_tolerance:
        print(total_voxels-voxels_within_tolerance)
        pytest.fail(f"Failed {config_id}")


def ibsi_ii_ph_ii_validation(ibsi_features, features):

    for raw_tag, feature_info in ibsi_features.items():
        tag = str(raw_tag)

        if tag in features:
            val = float(feature_info['consensus_value'])
            tol = float(feature_info['tolerance'])
            upper_boundary = val + tol
            lower_boundary = val - tol
            if not (lower_boundary <= features[tag] <= upper_boundary):
                pytest.fail(
                    f"Feature {tag} out of tolerance: {features[tag]} not in range ({lower_boundary}, {upper_boundary})")





@pytest.fixture()
def ct_phantom_image(ibsi_i_data_dir):
    image = Image()
    image.read_dicom_image(dicom_dir=str(ibsi_i_data_dir / 'dicom' / 'image'), modality='CT')
    return image


@pytest.fixture()
def ct_phantom_mask(ct_phantom_image, ibsi_i_data_dir):
    mask = Image()
    mask.read_dicom_mask(
        rtstruct_path=str(ibsi_i_data_dir / 'dicom' / 'mask' / 'DCM_RS_00060.dcm'),
        structure_name='GTV-1',
        image=ct_phantom_image
    )
    return mask


@pytest.fixture()
def res3d_1mm_image_spline(ct_phantom_image):
    preprocessing = Preprocessing(input_imaging_modality='CT',
                                  resample_resolution=1,
                                  resample_dimension='3D',
                                  interpolation_method='BSpline')
    res_image = preprocessing.resample(ct_phantom_image, image_type='image')

    return res_image


@pytest.fixture()
def res3d_1mm_mask_linear(ct_phantom_mask):
    preprocessing = Preprocessing(input_imaging_modality='CT',
                                  resample_resolution=1,
                                  resample_dimension='3D',
                                  interpolation_method='Linear',
                                  interpolation_threshold=.5)
    res_mask = preprocessing.resample(ct_phantom_mask, image_type='mask')

    return res_mask


@pytest.fixture()
def checkerboard_phantom():
    checkerboard = Image()
    checkerboard.read_nifti_image('tests/data/IBSI_II/Ph_I/nifti/checkerboard/image/checkerboard.nii.gz')
    return checkerboard


@pytest.fixture()
def impulse_phantom():
    impulse = Image()
    impulse.read_nifti_image('tests/data/IBSI_II/Ph_I/nifti/impulse/image/impulse.nii.gz')
    return impulse


@pytest.fixture()
def sphere_phantom():
    sphere = Image()
    sphere.read_nifti_image('tests/data/IBSI_II/Ph_I/nifti/sphere/image/sphere.nii.gz')
    return sphere


@pytest.mark.integration
def test_ibsi_ii_ph_i_1(ibsi_ii_data_dir, checkerboard_phantom, impulse_phantom):

    for config, params_and_images in {'1.a.1': ['constant', '3D', checkerboard_phantom, '1_a_1-ValidCRM.nii'],
                                      '1.a.2': ['nearest', '3D', checkerboard_phantom, '1_a_2-ValidCRM.nii'],
                                      '1.a.3': ['wrap', '3D', checkerboard_phantom, '1_a_3-ValidCRM.nii'],
                                      '1.a.4': ['reflect', '3D', checkerboard_phantom, '1_a_4-ValidCRM.nii'],
                                      '1.b.1': ['constant', '2D', impulse_phantom, '1_b_1-ValidCRM.nii']}.items():

        filtering = Filtering(filtering_method='Mean',
                              padding_type=params_and_images[0],
                              dimensionality=params_and_images[1],
                              support=15)

        _run_ph_i_case(filtering, params_and_images[-2], params_and_images[-1], config, ibsi_ii_data_dir)


@pytest.mark.integration
def test_ibsi_ii_ph_i_2(ibsi_ii_data_dir, checkerboard_phantom, impulse_phantom):

    for config, params_and_images in {'2.a': ['constant', '3D', 3.0, impulse_phantom, '2_a-ValidCRM.nii'],
                                      '2.b': ['reflect', '3D', 5.0, checkerboard_phantom, '2_b-ValidCRM.nii'],
                                      '2.c': ['reflect', '2D', 5.0, checkerboard_phantom, '2_c-ValidCRM.nii']}.items():

        filtering = Filtering(filtering_method='Laplacian of Gaussian',
                              padding_type=params_and_images[0],
                              dimensionality=params_and_images[1],
                              sigma_mm=params_and_images[2],
                              cutoff=4)

        _run_ph_i_case(filtering, params_and_images[-2], params_and_images[-1], config, ibsi_ii_data_dir)


@pytest.mark.integration
def test_ibsi_ii_ph_i_3(ibsi_ii_data_dir, checkerboard_phantom, impulse_phantom):

    for config, params_and_images in {'3.a.1': ['constant', '3D', 'E5L5S5', False, None, False, 0, impulse_phantom, '3_a_1-ValidCRM.nii'],
                                      '3.a.2': ['constant', '3D', 'E5L5S5', True, 'max', False, 0, impulse_phantom, '3_a_2-ValidCRM.nii'],
                                      '3.a.3': ['constant', '3D', 'E5L5S5', True, 'max', True, 7, impulse_phantom, '3_a_3-ValidCRM.nii'],
                                      '3.b.1': ['reflect', '3D', 'E3W5R5', False, None, False, 0, checkerboard_phantom, '3_b_1-ValidCRM.nii'],
                                      '3.b.2': ['reflect', '3D', 'E3W5R5', True, 'max', False, 0, checkerboard_phantom, '3_b_2-ValidCRM.nii'],
                                      '3.b.3': ['reflect', '3D', 'E3W5R5', True, 'max', True, 7, checkerboard_phantom, '3_b_3-ValidCRM.nii'],
                                      '3.c.1': ['reflect', '2D', 'L5S5', False, None, False, 0, checkerboard_phantom, '3_c_1-ValidCRM.nii'],
                                      '3.c.2': ['reflect', '2D', 'L5S5', True, 'max', False, 0, checkerboard_phantom, '3_c_2-ValidCRM.nii'],
                                      '3.c.3': ['reflect', '2D', 'L5S5', True, 'max', True, 7, checkerboard_phantom, '3_c_3-ValidCRM.nii'],
                                      }.items():

        filtering = Filtering(filtering_method='Laws Kernels',
                              padding_type=params_and_images[0],
                              dimensionality=params_and_images[1],
                              response_map=params_and_images[2],
                              rotation_invariance=params_and_images[3],
                              pooling=params_and_images[4],
                              energy_map=params_and_images[5],
                              distance=params_and_images[6])

        _run_ph_i_case(filtering, params_and_images[-2], params_and_images[-1], config, ibsi_ii_data_dir)


@pytest.mark.integration
@pytest.mark.parametrize(
    "config,padding,res_mm,sigma_mm,lambda_mm,gamma,theta,rot_inv,orth_planes,n_stds,phantom,truth_file",
    [
        ("4.a.1", "constant", 2.0, 10.0,   4.0, 1/2,     np.pi/3, False, False, 11, "impulse_phantom", "4_a_1-ValidCRM.nii"),
        ("4.a.2", "constant", 2.0, 10.0,   4.0, 1/2,     np.pi/4, True,  True,  11, "impulse_phantom", "4_a_2-ValidCRM.nii"),
        ("4.b.1", "reflect",  2.0, 20.0,   8.0, 5/2, 5*np.pi/4, False, False, None, "sphere_phantom",  "4_b_1-ValidCRM.nii"),
        ("4.b.2", "reflect",  2.0, 20.0,   8.0, 5/2,   np.pi/8, True,  True,  None, "sphere_phantom",  "4_b_2-ValidCRM.nii"),
    ],
    ids=lambda val, *_: val  # use the config string as the test‐id
)
def test_ibsi_ii_ph_i_4(
    config, padding, res_mm, sigma_mm, lambda_mm, gamma, theta,
    rot_inv, orth_planes, n_stds, phantom, truth_file,
    impulse_phantom, sphere_phantom, ibsi_ii_data_dir
):
    # pick the right fixture
    phantom_data = {'impulse_phantom': impulse_phantom,
                    'sphere_phantom': sphere_phantom}[phantom]

    filtering = Filtering(
        filtering_method='Gabor',
        padding_type=padding,
        res_mm=res_mm,
        sigma_mm=sigma_mm,
        lambda_mm=lambda_mm,
        gamma=gamma,
        theta=theta,
        rotation_invariance=rot_inv,
        orthogonal_planes=orth_planes,
        n_stds=n_stds
    )
    _run_ph_i_case(filtering, phantom_data, truth_file, config, ibsi_ii_data_dir)


@pytest.mark.integration
def test_ibsi_ii_ph_i_5(ibsi_ii_data_dir, impulse_phantom):

    for config, params_and_images in {'5.a.1': ['constant', 'LHL', False, impulse_phantom, '5_a_1-ValidCRM.nii'],
                                      '5.a.2': ['constant', 'LHL', True, impulse_phantom, '5_a_2-ValidCRM.nii'],
                                      }.items():
        filtering = Filtering(filtering_method='Wavelets',
                              wavelet_type="db2",
                              dimensionality='3D',
                              padding_type=params_and_images[0],
                              response_map=params_and_images[1],
                              decomposition_level=1,
                              rotation_invariance=params_and_images[2])

        _run_ph_i_case(filtering, params_and_images[-2], params_and_images[-1], config, ibsi_ii_data_dir)


@pytest.mark.integration
def test_ibsi_ii_ph_i_6(ibsi_ii_data_dir, sphere_phantom):

    for config, params_and_images in {'6.a.1': ['wrap', 'HHL', False, sphere_phantom, '6_a_1-ValidCRM.nii'],
                                      '6.a.2': ['wrap', 'HHL', True, sphere_phantom, '6_a_2-ValidCRM.nii'],
                                      }.items():
        filtering = Filtering(filtering_method='Wavelets',
                              wavelet_type="coif1",
                              dimensionality='3D',
                              padding_type=params_and_images[0],
                              response_map=params_and_images[1],
                              decomposition_level=1,
                              rotation_invariance=params_and_images[2])

        _run_ph_i_case(filtering, params_and_images[-2], params_and_images[-1], config, ibsi_ii_data_dir)


@pytest.mark.integration
def test_ibsi_ii_ph_i_7(ibsi_ii_data_dir, checkerboard_phantom):

    for config, params_and_images in {'7.a.1': ['reflect', 'LLL', False, checkerboard_phantom, '7_a_1-ValidCRM.nii'],
                                      '7.a.2': ['reflect', 'HHH', True, checkerboard_phantom, '7_a_2-ValidCRM.nii'],
                                      }.items():
        filtering = Filtering(filtering_method='Wavelets',
                              wavelet_type="haar",
                              dimensionality='3D',
                              padding_type=params_and_images[0],
                              response_map=params_and_images[1],
                              decomposition_level=2,
                              rotation_invariance=params_and_images[2])

        _run_ph_i_case(filtering, params_and_images[-2], params_and_images[-1], config, ibsi_ii_data_dir)


@pytest.mark.integration
def test_ibsi_ii_ph_ii_2a(ct_phantom_image, ct_phantom_mask):
    ibsi_features = ibsi_ii_feature_tolerances('2.A')

    filtering = Filtering(filtering_method='Mean',
                          padding_type='reflect',
                          dimensionality='2D',
                          support=5)

    filtered_image = filtering.apply_filter(ct_phantom_image)

    radiomics = Radiomics(aggr_dim='2D',
                          aggr_method='AVER',
                          intensity_range=[-1000, 400],
                          bin_size=25)

    radiomics.extract_features(image=ct_phantom_image, filtered_image=filtered_image, mask=ct_phantom_mask)
    ibsi_ii_ph_ii_validation(ibsi_features, radiomics.features_)


@pytest.mark.integration
def test_ibsi_ii_ph_ii_2b(res3d_1mm_image_spline, res3d_1mm_mask_linear):
    ibsi_features = ibsi_ii_feature_tolerances('2.B')

    filtering = Filtering(filtering_method='Mean',
                          padding_type='reflect',
                          dimensionality='3D',
                          support=5)

    filtered_image = filtering.apply_filter(res3d_1mm_image_spline)

    radiomics = Radiomics(aggr_dim='2D',
                          aggr_method='AVER',
                          intensity_range=[-1000, 400],
                          bin_size=25)

    radiomics.extract_features(image=res3d_1mm_image_spline, filtered_image=filtered_image, mask=res3d_1mm_mask_linear)
    ibsi_ii_ph_ii_validation(ibsi_features, radiomics.features_)


@pytest.mark.integration
def test_ibsi_ii_ph_ii_3a(ct_phantom_image, ct_phantom_mask):
    ibsi_features = ibsi_ii_feature_tolerances('3.A')

    filtering = Filtering(filtering_method='Laplacian of Gaussian',
                          padding_type='reflect',
                          dimensionality='2D',
                          sigma_mm=1.5,
                          cutoff=4)

    filtered_image = filtering.apply_filter(ct_phantom_image)

    radiomics = Radiomics(aggr_dim='2D',
                          aggr_method='AVER',
                          intensity_range=[-1000, 400],
                          bin_size=25)

    radiomics.extract_features(image=ct_phantom_image, filtered_image=filtered_image, mask=ct_phantom_mask)
    ibsi_ii_ph_ii_validation(ibsi_features, radiomics.features_)


@pytest.mark.integration
def test_ibsi_ii_ph_ii_3b(res3d_1mm_image_spline, res3d_1mm_mask_linear):
    ibsi_features = ibsi_ii_feature_tolerances('3.B')

    filtering = Filtering(filtering_method='Laplacian of Gaussian',
                          padding_type='reflect',
                          dimensionality='3D',
                          sigma_mm=1.5,
                          cutoff=4)

    filtered_image = filtering.apply_filter(res3d_1mm_image_spline)

    radiomics = Radiomics(aggr_dim='2D',
                          aggr_method='AVER',
                          intensity_range=[-1000, 400],
                          bin_size=25)

    radiomics.extract_features(image=res3d_1mm_image_spline, filtered_image=filtered_image, mask=res3d_1mm_mask_linear)
    ibsi_ii_ph_ii_validation(ibsi_features, radiomics.features_)


@pytest.mark.integration
def test_ibsi_ii_ph_ii_4a(ct_phantom_image, ct_phantom_mask):
    ibsi_features = ibsi_ii_feature_tolerances('4.A')

    filtering = Filtering(filtering_method='Laws Kernels',
                          padding_type='reflect',
                          dimensionality='2D',
                          response_map='L5E5',
                          rotation_invariance=True,
                          pooling="max",
                          energy_map=True,
                          distance=7)

    filtered_image = filtering.apply_filter(ct_phantom_image)

    radiomics = Radiomics(aggr_dim='2D',
                          aggr_method='AVER',
                          intensity_range=[-1000, 400],
                          bin_size=25)

    radiomics.extract_features(image=ct_phantom_image, filtered_image=filtered_image, mask=ct_phantom_mask)
    ibsi_ii_ph_ii_validation(ibsi_features, radiomics.features_)


@pytest.mark.integration
def test_ibsi_ii_ph_ii_4b(res3d_1mm_image_spline, res3d_1mm_mask_linear):
    ibsi_features = ibsi_ii_feature_tolerances('4.B')

    filtering = Filtering(filtering_method='Laws Kernels',
                          response_map="L5E5E5",
                          padding_type="reflect",
                          dimensionality="3D",
                          rotation_invariance=True,
                          pooling="max",
                          energy_map=True,
                          distance=7)

    filtered_image = filtering.apply_filter(res3d_1mm_image_spline)

    radiomics = Radiomics(aggr_dim='2D',
                          aggr_method='AVER',
                          intensity_range=[-1000, 400],
                          bin_size=25)

    radiomics.extract_features(image=res3d_1mm_image_spline, filtered_image=filtered_image, mask=res3d_1mm_mask_linear)
    ibsi_ii_ph_ii_validation(ibsi_features, radiomics.features_)

# Gabor
@pytest.mark.integration
def test_ibsi_ii_ph_ii_5a(ct_phantom_image, ct_phantom_mask):
    ibsi_features = ibsi_ii_feature_tolerances('5.A')

    filtering = Filtering(filtering_method='Gabor',
                          padding_type='reflect',
                          dimensionality='2D',
                          res_mm=0.977,
                          sigma_mm=5.0,
                          lambda_mm=2.0,
                          gamma=3/2,
                          theta=np.pi/8,
                          rotation_invariance=True,
                          orthogonal_planes=False)

    filtered_image = filtering.apply_filter(ct_phantom_image)

    radiomics = Radiomics(aggr_dim='2D',
                          aggr_method='AVER',
                          intensity_range=[-1000, 400],
                          bin_size=25)

    radiomics.extract_features(image=ct_phantom_image, filtered_image=filtered_image, mask=ct_phantom_mask)
    ibsi_ii_ph_ii_validation(ibsi_features, radiomics.features_)


@pytest.mark.integration
def test_ibsi_ii_ph_ii_5b(res3d_1mm_image_spline, res3d_1mm_mask_linear):
    ibsi_features = ibsi_ii_feature_tolerances('5.B')

    filtering = Filtering(filtering_method='Gabor',
                          padding_type="reflect",
                          dimensionality="3D",
                          res_mm=1.0,
                          sigma_mm=5.0,
                          lambda_mm=2.0,
                          gamma=3/2,
                          theta=np.pi/8,
                          rotation_invariance=True,
                          orthogonal_planes=True)

    filtered_image = filtering.apply_filter(res3d_1mm_image_spline)

    radiomics = Radiomics(aggr_dim='2D',
                          aggr_method='AVER',
                          intensity_range=[-1000, 400],
                          bin_size=25)

    radiomics.extract_features(image=res3d_1mm_image_spline, filtered_image=filtered_image, mask=res3d_1mm_mask_linear)
    ibsi_ii_ph_ii_validation(ibsi_features, radiomics.features_)

@pytest.mark.integration
def test_ibsi_ii_ph_ii_6a(ct_phantom_image, ct_phantom_mask):
    ibsi_features = ibsi_ii_feature_tolerances('6.A')

    filtering = Filtering(filtering_method='Wavelets',
                          wavelet_type="db3",
                          dimensionality='2D',
                          padding_type="reflect",
                          response_map="LH",
                          decomposition_level=1,
                          rotation_invariance=True)

    filtered_image = filtering.apply_filter(ct_phantom_image)

    radiomics = Radiomics(aggr_dim='2D',
                          aggr_method='AVER',
                          intensity_range=[-1000, 400],
                          bin_size=25)

    radiomics.extract_features(image=ct_phantom_image, filtered_image=filtered_image, mask=ct_phantom_mask)
    ibsi_ii_ph_ii_validation(ibsi_features, radiomics.features_)


@pytest.mark.integration
def test_ibsi_ii_ph_ii_6b(res3d_1mm_image_spline, res3d_1mm_mask_linear):
    ibsi_features = ibsi_ii_feature_tolerances('6.B')

    filtering = Filtering(filtering_method='Wavelets',
                          wavelet_type="db3",
                          dimensionality='3D',
                          padding_type="reflect",
                          response_map="LLH",
                          decomposition_level=1,
                          rotation_invariance=True)

    filtered_image = filtering.apply_filter(res3d_1mm_image_spline)

    radiomics = Radiomics(aggr_dim='2D',
                          aggr_method='AVER',
                          intensity_range=[-1000, 400],
                          bin_size=25)

    radiomics.extract_features(image=res3d_1mm_image_spline, filtered_image=filtered_image, mask=res3d_1mm_mask_linear)
    ibsi_ii_ph_ii_validation(ibsi_features, radiomics.features_)


@pytest.mark.integration
def test_ibsi_ii_ph_ii_7a(ct_phantom_image, ct_phantom_mask):
    ibsi_features = ibsi_ii_feature_tolerances('7.A')

    filtering = Filtering(filtering_method='Wavelets',
                          wavelet_type="db3",
                          dimensionality='2D',
                          padding_type="reflect",
                          response_map="HH",
                          decomposition_level=2,
                          rotation_invariance=True)

    filtered_image = filtering.apply_filter(ct_phantom_image)

    radiomics = Radiomics(aggr_dim='2D',
                          aggr_method='AVER',
                          intensity_range=[-1000, 400],
                          bin_size=25)

    radiomics.extract_features(image=ct_phantom_image, filtered_image=filtered_image, mask=ct_phantom_mask)
    ibsi_ii_ph_ii_validation(ibsi_features, radiomics.features_)


@pytest.mark.integration
def test_ibsi_ii_ph_ii_7b(res3d_1mm_image_spline, res3d_1mm_mask_linear):
    ibsi_features = ibsi_ii_feature_tolerances('7.B')

    filtering = Filtering(filtering_method='Wavelets', wavelet_type="db3",
                          dimensionality='3D',
                          padding_type="reflect",
                          response_map="HHH",
                          decomposition_level=2,
                          rotation_invariance=True)

    filtered_image = filtering.apply_filter(res3d_1mm_image_spline)

    radiomics = Radiomics(aggr_dim='2D',
                          aggr_method='AVER',
                          intensity_range=[-1000, 400],
                          bin_size=25)

    radiomics.extract_features(image=res3d_1mm_image_spline, filtered_image=filtered_image, mask=res3d_1mm_mask_linear)
    ibsi_ii_ph_ii_validation(ibsi_features, radiomics.features_)



"""Custom warnings and errors used across Z-Rad."""


class InvalidInputParametersError(ValueError):
    """Custom exception to indicate invalid input parameters."""


class DataStructureWarning(UserWarning):
    """Custom exception to indicate problems with input data structure."""


class DataStructureError(Exception):
    """Custom exception to indicate invalid input data structure."""



import pytest
import numpy as np
import SimpleITK as sitk
from zrad.preprocessing import Preprocessing
from zrad.image import Image

@pytest.mark.unit
def test_constructor_valid_inputs():
    preprocessing = Preprocessing(
        input_imaging_modality='CT',
        resample_resolution=2.0,
        resample_dimension='3D',
        interpolation_method='Linear',
        interpolation_threshold=0.5
    )
    assert preprocessing.input_imaging_modality == 'CT'
    assert preprocessing.resample_resolution == 2.0
    assert preprocessing.resample_dimension == '3D'
    assert preprocessing.interpolation_method == 'Linear'
    assert preprocessing.interpolation_threshold == 0.5


@pytest.mark.unit
@pytest.mark.parametrize("invalid_resolution", [0, -1, None, "abc"])
def test_constructor_invalid_resolution(invalid_resolution):
    with pytest.raises(ValueError) as exc_info:
        Preprocessing(
            input_imaging_modality='CT',
            resample_resolution=invalid_resolution,
            resample_dimension='2D',
            interpolation_method='Linear'
        )
    assert "must be a positive int or float" in str(exc_info.value)


@pytest.mark.unit
@pytest.mark.parametrize("invalid_dimension", ["1D", "4D", None, 123])
def test_constructor_invalid_dimension(invalid_dimension):
    with pytest.raises(ValueError) as exc_info:
        Preprocessing(
            input_imaging_modality='CT',
            resample_resolution=2.0,
            resample_dimension=invalid_dimension,
            interpolation_method='Linear'
        )
    assert "Resample dimension" in str(exc_info.value)
    assert "is not '2D' or '3D'." in str(exc_info.value)


@pytest.mark.unit
def test_get_interpolator_supported_method():
    interpolator = Preprocessing.get_interpolator('Linear')
    assert interpolator == sitk.sitkLinear

    interpolator = Preprocessing.get_interpolator('BSpline')
    assert interpolator == sitk.sitkBSpline


@pytest.mark.unit
def test_get_interpolator_unsupported_method():
    with pytest.raises(ValueError) as exc_info:
        Preprocessing.get_interpolator('SomeUnsupportedMethod')
    assert "is not supported" in str(exc_info.value)


@pytest.mark.unit
def test_calculate_resampled_origin():
    """
    Test that calculate_resampled_origin returns a float representing
    the new origin on a given axis.
    """
    initial_shape = (100, 100, 50)
    initial_spacing = (1.0, 1.0, 1.5)
    resulted_spacing = (2.0, 2.0, 1.5)
    initial_origin = (0.0, 0.0, 0.0)
    axis = 1  # Y-axis

    new_origin = Preprocessing.calculate_resampled_origin(
        initial_shape, initial_spacing, resulted_spacing, initial_origin, axis
    )

    # Basic sanity check: new origin should be within some reasonable range.
    assert isinstance(new_origin, float)


@pytest.mark.unit
def test_resample_image_2d():
    """
    Create a small 2D-like image (1 slice in the z-dimension) and check
    whether the Preprocessing class resamples it as expected.
    """
    # Dummy 2D data (we treat the z-dim as 1 to simulate a single slice)
    dummy_array = np.ones((1, 10, 10), dtype=np.float64)  # shape = (z, y, x)
    # Create an Image object with minimal metadata
    original_image = Image(
        array=dummy_array,
        origin=[0.0, 0.0, 0.0],
        spacing=[1.0, 1.0, 2.0],  # We can assume (x, y, z) or (y, x, z) as needed
        direction=[1, 0, 0, 0, 1, 0, 0, 0, 1],
        shape=(1, 10, 10)
    )

    preprocessing = Preprocessing(
        input_imaging_modality='CT',
        resample_resolution=2.0,
        resample_dimension='2D',   # We'll only change x and y spacing
        interpolation_method='Linear',
        interpolation_threshold=0.5
    )

    resampled_image = preprocessing.resample(original_image, image_type='image')

    # Verify dimensions, shape, spacing have changed for x,y but not for z
    assert resampled_image.shape[0] == 1  # z dimension unchanged
    assert resampled_image.spacing[2] == 2.0  # z spacing unchanged
    # x,y spacing should match 2.0 from the constructor
    assert pytest.approx(resampled_image.spacing[0]) == 2.0
    assert pytest.approx(resampled_image.spacing[1]) == 2.0

    # For a 10x10, upscaling spacing to 2.0 typically halves the shape dimension.
    # We might get 5x5 or 6x6, depending on ceiling. Just check it's smaller:
    assert resampled_image.shape[1] <= 10
    assert resampled_image.shape[2] <= 10


@pytest.mark.unit
def test_resample_mask_3d():
    """
    Create a small 3D mask image and verify thresholding is applied
    when resampling.
    """
    dummy_mask_array = np.array([[[0.4, 0.6],
                                  [1.0, 0.2]],
                                 [[0.9, 0.5],
                                  [0.2, 0.8]]],
                                dtype=np.float64)
    # shape = (2, 2, 2) in z,y,x order
    original_mask = Image(
        array=dummy_mask_array,
        origin=[0.0, 0.0, 0.0],
        spacing=[1.0, 1.0, 1.0],
        direction=[1, 0, 0, 0, 1, 0, 0, 0, 1],
        shape=dummy_mask_array.shape
    )

    preprocessing = Preprocessing(
        input_imaging_modality='CT',  # Modality is not as critical for masks
        resample_resolution=2.0,
        resample_dimension='3D',
        interpolation_method='Linear',
        interpolation_threshold=0.5
    )

    resampled_mask = preprocessing.resample(original_mask, image_type='mask')

    # Ensure the resampled array is in {0,1} due to the thresholding
    unique_values = np.unique(resampled_mask.array)
    assert all(val in [0, 1] for val in unique_values), "Mask should only have 0 or 1 after thresholding"



import sys
from datetime import datetime
from functools import lru_cache
from itertools import permutations

import cv2
import numpy as np
import pywt
from scipy import ndimage as ndi

from ..toolbox_logic import get_logger, handle_uncaught_exception, close_all_loggers

sys.excepthook = handle_uncaught_exception


class Mean:
    def __init__(self, padding_type, support, dimensionality):
        close_all_loggers()
        self.filter_logger_date_time = datetime.now().strftime("%Y-%m-%d_%H%M%S")
        self.filter_logger = get_logger(self.filter_logger_date_time+"_Mean_filter")

        self.type = 'Mean'

        if dimensionality in ['2D', '3D']:
            self.dimensionality = dimensionality
        else:
            self.filter_logger.error(f"Wrong dimensionality '{dimensionality}'. Available dimensions '2D' and '3D'.")
            raise ValueError(f"Wrong dimensionality '{dimensionality}'. Available dimensions '2D' and '3D'.")

        if isinstance(support, int):
            self.support = support
        else:
            self.filter_logger.error(f"Support should be int but '{type(support)}' detected.")
            raise ValueError(f"Support should be int but '{type(support)}' detected.")

        if padding_type in ['constant', 'nearest', 'wrap', 'reflect']:
            self.padding_type = padding_type
        else:
            self.filter_logger.error(f"Wrong padding type '{padding_type}'. "
                                     "Available padding types are: 'constant', 'nearest', 'wrap', and 'reflect'.")
            raise ValueError(f"Wrong padding type '{padding_type}'. "
                             "Available padding types are: 'constant', 'nearest', 'wrap', and 'reflect'.")

        self.filter_logger.debug(f"Defined {dimensionality} mean filter with support of {support}, "
                                 f"and {padding_type} padding type.")

    def apply(self, img):
        if self.dimensionality == "2D":
            filt_mat = np.ones([self.support, self.support])
            filt_mat = filt_mat / np.prod(filt_mat.shape)
            filtered_img = np.ones(img.shape)
            for i in range(img.shape[2]):
                filtered_img[:, :, i] = ndi.convolve(input=img[:, :, i], weights=filt_mat, mode=self.padding_type)
        elif self.dimensionality == "3D":
            filt_mat = np.ones([self.support, self.support, self.support])
            filt_mat = filt_mat / np.prod(filt_mat.shape)
            filtered_img = ndi.convolve(input=img, weights=filt_mat, mode=self.padding_type)
        else:
            filtered_img = None
        return filtered_img


class LoG:
    """LoG"""

    def __init__(self, padding_type, sigma_mm, cutoff, dimensionality):

        close_all_loggers()
        self.filter_logger_date_time = datetime.now().strftime("%Y-%m-%d_%H%M%S")
        self.filter_logger = get_logger(self.filter_logger_date_time+'_LoG_filter')

        self.type = 'Laplacian of Gaussian'

        if dimensionality in ['2D', '3D']:
            self.dimensionality = dimensionality
        else:
            self.filter_logger.error(f"Wrong dimensionality '{dimensionality}'. Available dimensions '2D' and '3D'.")
            raise ValueError(f"Wrong dimensionality '{dimensionality}'. Available dimensions '2D' and '3D'.")

        if padding_type in ['constant', 'nearest', 'wrap', 'reflect']:
            self.padding_type = padding_type
        else:
            self.filter_logger.error(f"Wrong padding type '{padding_type}'. "
                                     f"Available padding types are: 'constant', 'nearest', 'wrap', and 'reflect'.")
            raise ValueError(f"Wrong padding type '{padding_type}'. "
                             f"Available padding types are: 'constant', 'nearest', 'wrap', and 'reflect'.")

        if isinstance(sigma_mm, (int, float)):
            self.sigma_mm = sigma_mm
        else:
            self.filter_logger.error(f'Sigma (in mm) should be int or float but {type(sigma_mm)} detected.')
            raise ValueError(f'Sigma (in mm) should be int or float but {type(sigma_mm)} detected.')

        if isinstance(cutoff, (int, float)):
            self.cutoff = cutoff
        else:
            self.filter_logger.error(f'Cutoff should be int or float but {type(cutoff)} detected.')
            raise ValueError(f'Cutoff should be int or float but {type(cutoff)} detected.')

        self.padding_constant = 0.0
        self.res_mm = None

        self.filter_logger.debug(f"Defined {dimensionality} LoG filter with sigma {sigma_mm}, cutoff {cutoff}, "
                                 f"and {padding_type} padding type.")

    def apply(self, img):
        sigma = self.sigma_mm / self.res_mm
        if self.dimensionality == "3D":
            filtered_img = ndi.gaussian_laplace(img, sigma=sigma, mode=self.padding_type, cval=self.padding_constant,
                                                truncate=self.cutoff)
        elif self.dimensionality == "2D":
            filtered_img = np.nan * np.ones(img.shape)
            for i in range(img.shape[2]):
                filtered_img[:, :, i] = ndi.gaussian_laplace(img[:, :, i], sigma=sigma, mode=self.padding_type,
                                                             cval=self.padding_constant, truncate=self.cutoff)
        else:
            filtered_img = None
        return filtered_img


class Wavelets2D:
    """Wavelet filtering in 2D."""

    def __init__(self, wavelet_type, padding_type, response_map, decomposition_level, rotation_invariance=False):

        close_all_loggers()
        self.filter_logger_date_time = datetime.now().strftime("%Y-%m-%d_%H%M%S")
        self.filter_logger = get_logger(self.filter_logger_date_time+'_Wavelets2D')

        self.type = 'Wavelets'

        self.dimensionality = '2D'

        if padding_type in ['constant', 'nearest', 'wrap', 'reflect']:
            self.padding_type = padding_type
        else:
            self.filter_logger.error(f"Wrong padding type '{padding_type}'. "
                                     "Available padding types are: 'constant', 'nearest', 'wrap', and 'reflect'.")
            raise ValueError(f"Wrong padding type '{padding_type}'. "
                             "Available padding types are: 'constant', 'nearest', 'wrap', and 'reflect'.")

        if wavelet_type in ['db3', 'db2', 'coif1', 'haar']:
            self.wavelet_type = wavelet_type
        else:
            self.filter_logger.error(f"Wrong wavelet type '{wavelet_type}'. "
                                     "Available wavelet types are: 'db3', 'db2', 'coif1', 'haar'.")
            raise ValueError(f"Wrong wavelet type '{wavelet_type}'. "
                             "Available wavelet types are: 'db3', 'db2', 'coif1', 'haar'.")

        if decomposition_level in [1, 2]:
            self.decomposition_level = decomposition_level
        else:
            self.filter_logger.error(f"Wrong decomposition_level' {decomposition_level}'. "
                                     "Decomposition level should be integer. "
                                     "Available decomposition levels are: 1 and 2.")
            raise ValueError(f"Wrong decomposition_level' {decomposition_level}'. "
                             "Decomposition level should be integer. Available decomposition levels are: 1 and 2.")

        if response_map in ['LL', 'HL', 'LH', 'HH']:
            self.response_map = response_map
        else:
            self.filter_logger.error(f"Wrong response_map' {response_map}'. "
                                     "Available response_maps are: 'LL', 'HL', 'LH', 'HH'.")
            raise ValueError(f"Wrong response_map' {response_map}'. "
                             "Available response_maps are: 'LL', 'HL', 'LH', 'HH'.")

        if isinstance(rotation_invariance, bool):
            self.rotation_invariance = rotation_invariance
        else:
            self.filter_logger.error("Rotation Invariance should be "
                                     f"True or False but '{type(rotation_invariance)}' detected.")
            raise ValueError(f"Rotation Invariance should be True or False but '{type(rotation_invariance)}' detected.")

        self.filter_logger.debug(f"Defined 2D {wavelet_type} Wavelet filter with response map {response_map}, "
                                 f"decomposition level {decomposition_level}, "
                                 f"pseudo rotation invariance is {rotation_invariance}"
                                 f"and {padding_type} padding type.")

    def _get_kernel(self, response, decomposition_level=1):
        if response == "L":
            kernel = pywt.Wavelet(name=self.wavelet_type).filter_bank[0]
        elif response == "H":
            kernel = pywt.Wavelet(name=self.wavelet_type).filter_bank[1]
        else:
            kernel = None
        if decomposition_level == 2:
            kernel = [[e, 0] for e in kernel]
            kernel = [item for sublist in kernel for item in sublist]
        return kernel

    def _filter(self, img, x_filter, y_filter):
        filtered_img = ndi.convolve1d(img, x_filter, axis=1, mode=self.padding_type)
        filtered_img = ndi.convolve1d(filtered_img, y_filter, axis=0, mode=self.padding_type)
        return filtered_img

    def apply(self, img):
        if self.decomposition_level == 1:
            x_filter = self._get_kernel(self.response_map[0])
            y_filter = self._get_kernel(self.response_map[1])
            if self.rotation_invariance:
                final_image = np.zeros(img.shape)
                for i in range(img.shape[2]):
                    for k in range(4):
                        final_image[:, :, i] += np.rot90(self._filter(np.rot90(img[:, :, i], k=k, axes=(0, 1)),
                                                         x_filter, y_filter), k=k, axes=(1, 0))
                filtered_img = final_image / 4
            else:
                filtered_img = np.zeros(img.shape)
                for i in range(img.shape[2]):
                    filtered_img[:, :, i] = self._filter(img[:, :, i], x_filter, y_filter)
        elif self.decomposition_level == 2:
            l_filter = self._get_kernel("L")
            x_filter = self._get_kernel(self.response_map[0], decomposition_level=2)
            y_filter = self._get_kernel(self.response_map[1], decomposition_level=2)

            final_image = np.zeros(img.shape)
            for i in range(img.shape[2]):
                for k in range(4):
                    img_level0 = np.rot90(img[:, :, i], k=k, axes=(0, 1))  # original
                    img_level1 = self._filter(img_level0, l_filter, l_filter)
                    img_level2 = self._filter(img_level1, x_filter, y_filter)
                    final_image[:, :, i] += np.rot90(img_level2, k=k, axes=(1, 0))
            filtered_img = final_image / 4

        return filtered_img


class Wavelets3D:
    """Wavelet filtering."""

    def __init__(self, wavelet_type, padding_type, response_map, decomposition_level, rotation_invariance=False):

        close_all_loggers()
        self.filter_logger_date_time = datetime.now().strftime("%Y-%m-%d_%H%M%S")
        self.filter_logger = get_logger(self.filter_logger_date_time+'_Wavelets3D')

        self.type = 'Wavelets'

        self.dimensionality = '3D'

        if padding_type in ['constant', 'nearest', 'wrap', 'reflect']:
            self.padding_type = padding_type
        else:
            self.filter_logger.error(f"Wrong padding type '{padding_type}'. "
                                     "Available padding types are: 'constant', 'nearest', 'wrap', and 'reflect'.")
            raise ValueError(f"Wrong padding type '{padding_type}'. "
                             "Available padding types are: 'constant', 'nearest', 'wrap', and 'reflect'.")

        if wavelet_type in ['db3', 'db2', 'coif1', 'haar']:
            self.wavelet_type = wavelet_type
        else:
            self.filter_logger.error(f"Wrong wavelet type '{wavelet_type}'. "
                                     "Available wavelet types are: 'db3', 'db2', 'coif1', 'haar'.")
            raise ValueError(f"Wrong wavelet type '{wavelet_type}'. "
                             "Available wavelet types are: 'db3', 'db2', 'coif1', 'haar'.")

        if decomposition_level in [1, 2]:
            self.decomposition_level = decomposition_level
        else:
            self.filter_logger.error(f"Wrong decomposition_level' {decomposition_level}'. "
                                     "Decomposition level should be integer. "
                                     "Available decomposition levels are: 1 and 2.")
            raise ValueError(f"Wrong decomposition_level' {decomposition_level}'. "
                             "Decomposition level should be integer. Available decomposition levels are: 1 and 2.")

        if response_map in ['LLL', 'LLH', 'LHL', 'HLL', 'LHH', 'HHL', 'HLH', 'HHH']:
            self.response_map = response_map
        else:
            self.filter_logger.error(f"Wrong response_map' {response_map}'. "
                                     "Available response_maps are: "
                                     "'LLL', 'LLH', 'LHL', 'HLL', 'LHH', 'HHL', 'HLH', 'HHH'.")
            raise ValueError(f"Wrong response_map' {response_map}'. "
                             "Available response_maps are: 'LLL', 'LLH', 'LHL', 'HLL', 'LHH', 'HHL', 'HLH', 'HHH'.")

        if isinstance(rotation_invariance, bool):
            self.rotation_invariance = rotation_invariance
        else:
            self.filter_logger.error("Rotation Invariance should be "
                                     f"True or False but '{type(rotation_invariance)}' detected.")
            raise ValueError(f"Rotation Invariance should be True or False but '{type(rotation_invariance)}' detected.")

        self.pooling = None

        self.filter_logger.debug(f"Defined 3D {wavelet_type} Wavelet filter with response map {response_map}, "
                                 f"decomposition level {decomposition_level}, "
                                 f"pseudo rotation invariance is {rotation_invariance}"
                                 f"and {padding_type} padding type.")

    def _get_kernel(self, response, decomposition_level=1):
        if response == "L":
            kernel = pywt.Wavelet(name=self.wavelet_type).filter_bank[0]
        elif response == "H":
            kernel = pywt.Wavelet(name=self.wavelet_type).filter_bank[1]
        else:
            kernel = None
        if decomposition_level == 2:
            kernel = [[e, 0] for e in kernel]
            kernel = [item for sublist in kernel for item in sublist]
        return kernel

    def _filter(self, img, x_filter, y_filter, z_filter):
        filtered_img = ndi.convolve1d(img, x_filter, axis=1, mode=self.padding_type)
        filtered_img = ndi.convolve1d(filtered_img, y_filter, axis=0, mode=self.padding_type)
        filtered_img = ndi.convolve1d(filtered_img, z_filter, axis=2, mode=self.padding_type)
        return filtered_img

    def apply(self, img):
        if self.decomposition_level == 1:
            x_filter = self._get_kernel(self.response_map[0])
            y_filter = self._get_kernel(self.response_map[1])
            z_filter = self._get_kernel(self.response_map[2])
            if self.rotation_invariance:
                final_image = np.zeros(img.shape)
                kernels_permutation = [(x_filter, y_filter, z_filter),
                                       (z_filter, x_filter, y_filter),
                                       (y_filter, z_filter, x_filter)]
                for kernels in kernels_permutation:
                    final_image += self._filter(img, kernels[0], kernels[1], kernels[2])
                    final_image += self._filter(img[::-1, :, :], kernels[0], kernels[1], kernels[2])[::-1, :, :]
                    final_image += self._filter(img[:, ::-1, :], kernels[0], kernels[1], kernels[2])[:, ::-1, :]
                    final_image += self._filter(img[:, :, ::-1], kernels[0], kernels[1], kernels[2])[:, :, ::-1]
                    final_image += self._filter(img[::-1, ::-1, :], kernels[0], kernels[1], kernels[2])[::-1, ::-1, :]
                    final_image += self._filter(img[::-1, :, ::-1], kernels[0], kernels[1], kernels[2])[::-1, :, ::-1]
                    final_image += self._filter(img[:, ::-1, ::-1], kernels[0], kernels[1], kernels[2])[:, ::-1, ::-1]
                    final_image += self._filter(img[::-1, ::-1, ::-1], kernels[0], kernels[1], kernels[2])[::-1, ::-1, ::-1]
                filtered_img = final_image / (8 * len(kernels_permutation))
            else:
                filtered_img = self._filter(img, x_filter, y_filter, z_filter)
        else:
            # First, do low pass LLL
            x_filter = self._get_kernel("L")
            y_filter = self._get_kernel("L")
            z_filter = self._get_kernel("L")
            kernels_permutation = [(x_filter, y_filter, z_filter),
                                   (z_filter, x_filter, y_filter),
                                   (y_filter, z_filter, x_filter)]
            level1_responses = list()
            for kernels in kernels_permutation:
                level1_responses.append(self._filter(img, kernels[0], kernels[1], kernels[2]))
                level1_responses.append(self._filter(img[::-1, :, :], kernels[0], kernels[1], kernels[2])[::-1, :, :])
                level1_responses.append(self._filter(img[:, ::-1, :], kernels[0], kernels[1], kernels[2])[:, ::-1, :])
                level1_responses.append(self._filter(img[:, :, ::-1], kernels[0], kernels[1], kernels[2])[:, :, ::-1])
                level1_responses.append(self._filter(img[::-1, ::-1, :], kernels[0], kernels[1], kernels[2])[::-1, ::-1, :])
                level1_responses.append(self._filter(img[::-1, :, ::-1], kernels[0], kernels[1], kernels[2])[::-1, :, ::-1])
                level1_responses.append(self._filter(img[:, ::-1, ::-1], kernels[0], kernels[1], kernels[2])[:, ::-1, ::-1])
                level1_responses.append(self._filter(img[::-1, ::-1, ::-1], kernels[0], kernels[1], kernels[2])[::-1, ::-1, ::-1])

            # Now do 2nd level filtering
            x_filter = self._get_kernel(self.response_map[0], decomposition_level=2)
            y_filter = self._get_kernel(self.response_map[1], decomposition_level=2)
            z_filter = self._get_kernel(self.response_map[2], decomposition_level=2)
            final_image = np.zeros(img.shape)
            kernels_permutation = [(x_filter, y_filter, z_filter),
                                   (z_filter, x_filter, y_filter),
                                   (y_filter, z_filter, x_filter)]
            for kernels in kernels_permutation:
                final_image += self._filter(level1_responses[0], kernels[0], kernels[1], kernels[2])
                final_image += self._filter(level1_responses[1][::-1, :, :], kernels[0], kernels[1], kernels[2])[::-1, :, :]
                final_image += self._filter(level1_responses[2][:, ::-1, :], kernels[0], kernels[1], kernels[2])[:, ::-1, :]
                final_image += self._filter(level1_responses[3][:, :, ::-1], kernels[0], kernels[1], kernels[2])[:, :, ::-1]
                final_image += self._filter(level1_responses[4][::-1, ::-1, :], kernels[0], kernels[1], kernels[2])[::-1, ::-1, :]
                final_image += self._filter(level1_responses[5][::-1, :, ::-1], kernels[0], kernels[1], kernels[2])[::-1, :, ::-1]
                final_image += self._filter(level1_responses[6][:, ::-1, ::-1], kernels[0], kernels[1], kernels[2])[:, ::-1, ::-1]
                final_image += self._filter(level1_responses[7][::-1, ::-1, ::-1], kernels[0], kernels[1], kernels[2])[::-1, ::-1, ::-1]
            filtered_img = final_image / (8 * len(kernels_permutation))
        return filtered_img


class Laws:
    """Laws2"""

    def __init__(self, response_map, padding_type, distance, energy_map, dimensionality,
                 rotation_invariance=False, pooling=None):

        close_all_loggers()
        self.filter_logger_date_time = datetime.now().strftime("%Y-%m-%d_%H%M%S")
        self.filter_logger = get_logger(self.filter_logger_date_time+'_Laws_kernels')

        self.type = 'Laws Kernels'

        if dimensionality in ['2D', '3D']:
            self.dimensionality = dimensionality
        else:
            self.filter_logger.error(f"Wrong dimensionality '{dimensionality}'. Available dimensions '2D' and '3D'.")
            raise ValueError(f"Wrong dimensionality '{dimensionality}'. Available dimensions '2D' and '3D'.")

        if padding_type in ['constant', 'nearest', 'wrap', 'reflect']:
            self.padding_type = padding_type
        else:
            self.filter_logger.error(f"Wrong padding type '{padding_type}'. "
                                     "Available padding types are: 'constant', 'nearest', 'wrap', and 'reflect'.")
            raise ValueError(f"Wrong padding type '{padding_type}'. "
                             "Available padding types are: 'constant', 'nearest', 'wrap', and 'reflect'.")

        if isinstance(distance, int):
            self.distance = distance
        else:
            self.filter_logger.error(f"Distance should be 'int' but '{type(distance)}' detected.")
            raise ValueError(f"Distance should be 'int' but '{type(distance)}' detected.")

        if isinstance(energy_map, bool):
            self.energy_map = energy_map
        else:
            self.filter_logger.error('Energy map can be only True or False.')
            raise ValueError('Energy map can be only True or False.')

        if isinstance(rotation_invariance, bool):
            self.rotation_invariance = rotation_invariance
        else:
            self.filter_logger.error("Rotation Invariance should be "
                                     f"True or False but '{type(rotation_invariance)}' detected.")
            raise ValueError(f"Rotation Invariance should be True or False but '{type(rotation_invariance)}' detected.")

        self.response_map = response_map
        self.pooling = pooling

        self.filter_logger.debug(f"Defined {dimensionality} Laws Kernels filter with energy map is {energy_map}, "
                                 f"response map {response_map}, distance {distance} "
                                 f"pseudo rotation invariance is {rotation_invariance}, pooling {pooling},"
                                 f"and {padding_type} padding type.")

    def _get_kernel(self, l_type, support):
        if l_type == "L":
            if support == 3:
                return 1 / np.sqrt(6) * np.array([1, 2, 1])
            elif support == 5:
                return 1 / np.sqrt(70) * np.array([1, 4, 6, 4, 1])
        elif l_type == "E":
            if support == 3:
                return 1 / np.sqrt(2) * np.array([-1, 0, 1])
            elif support == 5:
                return 1 / np.sqrt(10) * np.array([-1, -2, 0, 2, 1])
        elif l_type == "S":
            if support == 3:
                return 1 / np.sqrt(6) * np.array([-1, 2, -1])
            elif support == 5:
                return 1 / np.sqrt(6) * np.array([-1, 0, 2, 0, -1])
        elif l_type == "W":
            if support == 5:
                return 1 / np.sqrt(10) * np.array([-1, 2, 0, -2, 1])
        elif l_type == "R":
            if support == 5:
                return 1 / np.sqrt(70) * np.array([1, -4, 6, -4, 1])

    def _get_response_maps(self):
        parts = [self.response_map[i:i + 2] for i in range(0, len(self.response_map), 2)]
        return [''.join(e) for e in permutations(parts)]

    def _filter(self, img, response_map):
        if self.dimensionality == "3D":
            x_filt = self._get_kernel(response_map[0], int(response_map[1]))
            y_filt = self._get_kernel(response_map[2], int(response_map[3]))
            z_filt = self._get_kernel(response_map[4], int(response_map[5]))

            filtered_img = ndi.convolve1d(img, x_filt, axis=1, mode=self.padding_type)
            filtered_img = ndi.convolve1d(filtered_img, y_filt, axis=0, mode=self.padding_type)
            filtered_img = ndi.convolve1d(filtered_img, z_filt, axis=2, mode=self.padding_type)
        elif self.dimensionality == "2D":
            x_filt = self._get_kernel(response_map[0], int(response_map[1]))
            y_filt = self._get_kernel(response_map[2], int(response_map[3]))

            filtered_img = ndi.convolve1d(img, x_filt, axis=1, mode=self.padding_type)
            filtered_img = ndi.convolve1d(filtered_img, y_filt, axis=0, mode=self.padding_type)
        else:
            filtered_img = None
        return filtered_img

    def apply(self, img):
        final_image = None
        if self.rotation_invariance:
            response_maps = self._get_response_maps()

            if self.pooling == "avg":
                # average pooling of Laws filtering was not tested in IBSI 2.
                final_image = np.nan * np.ones(img.shape)
                for response_map in response_maps:
                    final_image += self._filter(img, response_map)
                    final_image += self._filter(img[::-1, :, :], response_map)[::-1, :, :]
                    final_image += self._filter(img[:, ::-1, :], response_map)[:, ::-1, :]
                    final_image += self._filter(img[:, :, ::-1], response_map)[:, :, ::-1]
                    final_image += self._filter(img[::-1, ::-1, :], response_map)[::-1, ::-1, :]
                    final_image += self._filter(img[::-1, :, ::-1], response_map)[::-1, :, ::-1]
                    final_image += self._filter(img[:, ::-1, ::-1], response_map)[:, ::-1, ::-1]
                    final_image += self._filter(img[::-1, ::-1, ::-1], response_map)[::-1, ::-1, ::-1]
                final_image = final_image / 24  # shouldn't it rather be len(response_maps) * 8 = 48?

            elif self.pooling == "max":
                final_image = -np.inf * np.ones(img.shape)
                for response_map in response_maps:
                    final_image = np.maximum(final_image, self._filter(img, response_map))
                    final_image = np.maximum(final_image, self._filter(img[::-1, :, :], response_map)[::-1, :, :])
                    final_image = np.maximum(final_image, self._filter(img[:, ::-1, :], response_map)[:, ::-1, :])
                    final_image = np.maximum(final_image, self._filter(img[:, :, ::-1], response_map)[:, :, ::-1])
                    final_image = np.maximum(final_image, self._filter(img[::-1, ::-1, :], response_map)[::-1, ::-1, :])
                    final_image = np.maximum(final_image, self._filter(img[::-1, :, ::-1], response_map)[::-1, :, ::-1])
                    final_image = np.maximum(final_image, self._filter(img[:, ::-1, ::-1], response_map)[:, ::-1, ::-1])
                    final_image = np.maximum(final_image,
                                             self._filter(img[::-1, ::-1, ::-1], response_map)[::-1, ::-1, ::-1])
        else:
            final_image = self._filter(img, self.response_map)

        if self.energy_map:
            final_image = self._get_energy_map(final_image)

        return final_image

    def _get_energy_map(self, img):
        if self.dimensionality == "2D":
            filt_mat = np.ones([2 * self.distance + 1, 2 * self.distance + 1])
            filt_mat = filt_mat / np.prod(filt_mat.shape)
            energy_map = np.nan * np.ones(img.shape)
            for i in range(img.shape[2]):
                energy_map[:, :, i] = ndi.convolve(input=np.abs(img[:, :, i]), weights=filt_mat, mode='reflect')
        elif self.dimensionality == "3D":
            filt_mat = np.ones([2 * self.distance + 1, 2 * self.distance + 1, 2 * self.distance + 1])
            filt_mat = filt_mat / np.prod(filt_mat.shape)
            energy_map = ndi.convolve(input=np.abs(img), weights=filt_mat, mode='reflect')
        else:
            energy_map = None
        return energy_map


class Gabor:
    """
    Gabor filter bank for 3D volumes, supporting single-θ or rotation-invariant responses.

    This class constructs and applies 2D Gabor filters slice‑wise along any orthogonal plane
    of a 3D image, caching filter kernels and optionally averaging over orientations and planes.
    """
    _PADDING_MAP = {
        'reflect': cv2.BORDER_REFLECT,
        'mirror': cv2.BORDER_REFLECT_101,
        'constant': cv2.BORDER_CONSTANT,
        'nearest': cv2.BORDER_REPLICATE,
        'wrap': cv2.BORDER_WRAP,
    }

    def __init__(self,
                 padding_type: str,
                 res_mm: float,
                 sigma_mm: float,
                 lambda_mm: float,
                 gamma: float,
                 theta: float,
                 rotation_invariance: bool = False,
                 orthogonal_planes: bool = False,
                 n_stds: float = None):
        """
        Initialize a Gabor filter instance.

        Parameters
        ----------
        padding_type
            One of {'reflect', 'mirror', 'constant', 'nearest', 'wrap'}; maps to OpenCV borderType.
        res_mm
            Image resolution in millimetres per pixel.
        sigma_mm
            Gaussian envelope standard deviation in millimetres.
        lambda_mm
            Wavelength of the sinusoidal component in millimetres.
        gamma
            Spatial aspect ratio (controls ellipticity).
        theta
            Angular frequency or step increment (in radians) for orientation(s).
        rotation_invariance
            If True, average responses over a full orientation bank.
        orthogonal_planes
            If True and rotation_invariance, include (0,2) and (1,2) planes as well.
        n_stds
            Number of σ (in pixels) to span for the kernel half‑width; if None, defaults internally.

        Raises
        ------
        ValueError
            If `padding_type` is not a recognized key.
        """

        try:
            self._border = self._PADDING_MAP[padding_type]
        except KeyError:
            raise ValueError(f"padding_type must be one of {list(self._PADDING_MAP)}, "
                             f"got {padding_type!r}")
        self.rotation_invariance = rotation_invariance
        self.res_mm = res_mm
        self.theta = theta
        self.gamma = gamma
        self.lambda_mm = lambda_mm
        self.sigma_mm = sigma_mm
        self.padding_type = padding_type
        self.orthogonal_planes = orthogonal_planes
        self.n_stds = n_stds

    # -------------------------------------------------------------------------
    # 1.  KERNEL FACTORY WITH CACHING
    # -------------------------------------------------------------------------
    @lru_cache(maxsize=128)  # theta‑specific, shape‑specific memoisation
    def _make_kernels(self, theta, ksize):
        """
        Build and cache real & imaginary 2D Gabor kernels in pixel units.

        This method enforces an odd `ksize`, converts `sigma_mm` and `lambda_mm`
        to pixels via `res_mm`, and returns float32 kernels for `psi=0` and `psi=π/2`.

        Parameters
        ----------
        theta
            Filter orientation angle in radians.
        ksize
            Proposed kernel dimension (will be bumped to odd if even).

        Returns
        -------
        kern_real : np.ndarray
            Real (cosine-phase) Gabor kernel of shape (ksize, ksize).
        kern_imag : np.ndarray
            Imaginary (sine-phase) Gabor kernel of shape (ksize, ksize).
        """
        if ksize % 2 == 0:
            ksize += 1
        kern_real = cv2.getGaborKernel(
            (ksize, ksize), self.sigma_mm / self.res_mm,
            theta, self.lambda_mm / self.res_mm,
            self.gamma, 0, ktype=cv2.CV_32F)
        kern_imag = cv2.getGaborKernel(
            (ksize, ksize), self.sigma_mm / self.res_mm,
            theta, self.lambda_mm / self.res_mm,
            self.gamma, np.pi / 2, ktype=cv2.CV_32F)
        return kern_real, kern_imag

    # -------------------------------------------------------------------------
    # 2.  INNER WORKHORSE – single θ, single 2‑D plane
    # -------------------------------------------------------------------------
    def _filter(self, img, theta, plane2d=(0, 1)):
        """
        Compute the magnitude response for one θ on a single 2D plane within a 3D image.

        This method:
        1. Reorders axes so `plane2d` maps to (y,x).
        2. Selects a kernel size via `n_stds` or default.
        3. Convolves each slice with the cached Gabor kernels.
        4. Returns the magnitude response and restores the original axis order.

        Parameters
        ----------
        img
            Input 3D image array.
        theta
            Orientation angle in radians.
        plane2d
            Tuple of two axes to filter (e.g., (0,1), (0,2), (1,2)).

        Returns
        -------
        filtered_img : np.ndarray
            3D volume of the same shape as `img`, containing the magnitude response.
        """
        # --- a. bring the target plane to (y, x, z) order without copying -----
        axes = list(plane2d) + [i for i in range(3) if i not in plane2d]
        img_view = np.transpose(img, axes).astype(np.float32, copy=False)

        # --- b. pick a compact kernel  ------
        if self.n_stds is None:
            ksize = int(np.ceil(7 * (self.sigma_mm / self.res_mm))) | 1
        else:
            ksize = int(np.ceil(self.n_stds * (self.sigma_mm / self.res_mm))) | 1
        kern_r, kern_i = self._make_kernels(theta, ksize)

        # --- c. convolve each slice with OpenCV’s highly vectorised filter ----
        out = np.empty_like(img_view)
        for z in range(img_view.shape[2]):
            slice_ = img_view[:, :, z]
            # ‑ OpenCV is > 2× faster than scipy.ndimage.convolve on small kernels
            out_r = cv2.filter2D(slice_, -1, kern_r, borderType=self._border)
            out_i = cv2.filter2D(slice_, -1, kern_i, borderType=self._border)
            out[:, :, z] = np.hypot(out_r, out_i)

        # --- d. put axes back ------
        return np.transpose(out, np.argsort(axes))

    # -------------------------------------------------------------------------
    # 3.  PUBLIC WRAPPER – multiple θ and planes
    # -------------------------------------------------------------------------
    def filter(self, img):
        """
        Apply the Gabor filter to a 3D image, optionally with rotation invariance.

        If `rotation_invariance=True`, this computes responses at orientations
        from 0 to 2π with step `theta` and averages over the specified plane(s).
        By default, only the (0,1) plane is used; if `self.orthogonal_planes=True`,
        the method also includes the (0,2) and (1,2) planes in the averaging.

        Parameters
        ----------
        img : np.ndarray
            Input 3D image array.

        Returns
        -------
        result : np.ndarray
            Filtered 3D image of the same shape as `img`.
        """
        if self.rotation_invariance:
            thetas = np.arange(0, 2 * np.pi, self.theta, dtype=np.float32)
            planes = [(0, 1), (0, 2), (1, 2)] if self.orthogonal_planes else [(0, 1)]
            resp = [self._filter(img, th, pl) for th in thetas for pl in planes]
            return np.mean(resp, axis=0, dtype=np.float32)
        return self._filter(img, self.theta)



import copy
import os
import warnings
from datetime import datetime

import SimpleITK as sitk
import numpy as np
import pydicom
from pydicom.errors import InvalidDicomError
from skimage import draw

from .exceptions import DataStructureWarning, DataStructureError


def parse_time(time_str):
    """
    Parse a time string into a datetime object using various possible formats.

    Args:
        time_str (str or bytes): The time string to parse.

    Returns:
        datetime: Parsed datetime object.

    Raises:
        ValueError: If the time string does not match any expected formats.
    """
    if isinstance(time_str, bytes):
        time_str = time_str.decode('utf-8').strip()

    for fmt in ('%H%M%S.%f', '%H%M%S', '%Y%m%d%H%M%S.%f', '%Y%m%d%H%M%S'):
        try:
            return datetime.strptime(time_str, fmt)
        except ValueError:
            continue
        except TypeError:
            continue
    raise ValueError(f"Time data '{time_str}' does not match expected formats")


class Image:
    def __init__(self, array=None, origin=None, spacing=None, direction=None, shape=None):
        self.sitk_image = None
        self.array = array
        self.origin = origin
        self.spacing = spacing
        self.direction = direction
        self.shape = shape

    def copy(self):
        return Image(
            array=copy.deepcopy(self.array),
            origin=copy.deepcopy(self.origin),
            spacing=copy.deepcopy(self.spacing),
            direction=copy.deepcopy(self.direction),
            shape=copy.deepcopy(self.shape)
        )

    def save_as_nifti(self, output_path):
        if not os.path.exists(os.path.dirname(output_path)):
            os.makedirs(os.path.dirname(output_path))
        img = sitk.GetImageFromArray(self.array)
        img.SetOrigin(self.origin)
        img.SetSpacing(self.spacing)
        img.SetDirection(self.direction)
        sitk.WriteImage(img, output_path)

    def read_nifti_image(self, image_path):
        sitk_reader = sitk.ImageFileReader()
        sitk_reader.SetImageIO("NiftiImageIO")
        sitk_reader.SetFileName(image_path)
        image = sitk_reader.Execute()
        array = sitk.GetArrayFromImage(image)
        self.sitk_image = image
        self.array = array.astype(np.float64)
        self.origin = image.GetOrigin()
        self.spacing = np.array(image.GetSpacing())
        self.direction = image.GetDirection()
        self.shape = image.GetSize()

    def read_nifti_mask(self, image, mask_path):
        sitk_reader = sitk.ImageFileReader()
        sitk_reader.SetImageIO("NiftiImageIO")
        sitk_reader.SetFileName(mask_path)
        mask = sitk_reader.Execute()
        array = sitk.GetArrayFromImage(mask)
        self.sitk_image = image
        self.array = array.astype(np.float64)
        self.origin = image.origin
        self.spacing = image.spacing
        self.direction = image.direction
        self.shape = image.shape

    def read_dicom_image(self, dicom_dir, modality):
        dicom_files = get_dicom_files(directory=dicom_dir, modality=modality)
        if len(dicom_files) == 0:
            raise DataStructureError(f"No {modality} data found in {dicom_dir}. Patient skipped.")
        if modality in ["CT", "MRI", "PET"]:
            validate_z_spacing(dicom_files)
        if modality in ["CT", "MRI", "PET", "MG"]:
            image = process_dicom_series(dicom_files)
        if modality == 'PET':
            validate_pet_dicom_tags(dicom_files)
            image = apply_suv_correction(dicom_files, image)
        if modality == 'RTDOSE':
            file_path = dicom_files[0]['file_path']
            image = self.read_dicom_dose(file_path)
        if image:
            array = sitk.GetArrayFromImage(image)
            self.sitk_image = image
            self.array = array.astype(np.float64)
            self.origin = image.GetOrigin()
            self.spacing = np.array(image.GetSpacing())
            self.direction = image.GetDirection()
            self.shape = image.GetSize()

    def read_dicom_mask(self, rtstruct_path, structure_name, image):
        mask = extract_dicom_mask(rtstruct_path, structure_name, image.sitk_image)
        self.array = mask.array
        self.origin = mask.origin
        self.spacing = mask.spacing
        self.direction = mask.direction
        self.shape = mask.shape

    def read_dicom_dose(self, rtdose_path):
        ds = pydicom.dcmread(rtdose_path)
        if ds.DoseUnits != 'GY':
            raise DataStructureError(f"Only dose in Gy is supported. Provided {ds.DoseUnits}. Patient skipped")
        if ds.DoseType != 'PHYSICAL':
            raise DataStructureError(f"Only physical dose is supported. Provided {ds.DoseType}. Patient skipped.")
        raw_dose_image = sitk.ReadImage(rtdose_path)
        dose_array = sitk.GetArrayFromImage(raw_dose_image) * ds.DoseGridScaling
        dose_image = sitk.GetImageFromArray(
            dose_array)  # dose_array is a NumPy array that contains the dose values (after applying DoseGridScaling). This line converts dose_array back into a SITK image.
        dose_image.SetOrigin(
            raw_dose_image.GetOrigin())  # Copies the origin (spatial reference of the image in the coordinate system) from the original DICOM image (raw_dose_image) to dose_image.
        dose_image.SetSpacing(
            raw_dose_image.GetSpacing())  # Ensures that the voxel spacing (physical size of each pixel in millimeters) is retained from the original dose DICOM file.
        dose_image.SetDirection(
            raw_dose_image.GetDirection())  # Copies the image orientation (how the image is aligned in 3D space) from the original dose DICOM file.

        return dose_image

def remove_duplicate_slices(dicom_files_info):
    """
    Remove duplicate slices with identical ImagePositionPatient (IPP).

    Keeps the first occurrence of each IPP (in the order coming from
    GetGDCMSeriesFileNames, which is already geometrically sorted).
    """
    cleaned = []
    seen_ipps = set()
    duplicates = 0

    for info in dicom_files_info:
        ds = info['ds']
        
        if "ImagePositionPatient" in ds:
            ipp = tuple(map(float, ds.ImagePositionPatient))
        else:
            ipp = None


        if ipp is None:
            cleaned.append(info)
            continue

        if ipp in seen_ipps:
            duplicates += 1
            continue

        seen_ipps.add(ipp)
        cleaned.append(info)
    if duplicates > 0:
        warnings.warn(
            f"Removed {duplicates} duplicate slice(s) with identical ImagePositionPatient.",
            DataStructureWarning
        )

    return cleaned

def sort_by_geometric_position(dicom_files_info):
    """
    Sort slices by physical position using ImageOrientationPatient + ImagePositionPatient
    (distance along the slice normal).
    """

    def slice_distance(ds):
        iop = np.array(ds.ImageOrientationPatient, dtype=float)
        row = iop[:3]
        col = iop[3:]
        normal = np.cross(row, col)
        ipp = np.array(ds.ImagePositionPatient, dtype=float)
        return float(np.dot(ipp, normal))

    return sorted(dicom_files_info, key=lambda x: slice_distance(x['ds']))

def get_dicom_files(directory, modality):
    modality_dicom = modality_mapping(modality)
    dicom_files_info = []
    if modality_dicom in ['CT', 'PT', 'MR']:
        reader = sitk.ImageSeriesReader()
        series_IDs = reader.GetGDCMSeriesIDs(directory)
        selected_series = None
        for sid in series_IDs:
            files = reader.GetGDCMSeriesFileNames(directory, sid)
            dcm = pydicom.dcmread(os.path.join(directory, files[0]), stop_before_pixels=True)
            if dcm.Modality == modality_dicom:
                selected_series = sid
                break
        if selected_series is None:
            raise DataStructureError(f"No {modality_dicom} series found for {directory}. Patient skipped")
        all_files = reader.GetGDCMSeriesFileNames(directory, selected_series)
    else:
        all_files = [os.path.join(directory, i) for i in os.listdir(directory)]
    acquisition_numbers = []
    for file_path in all_files:
        try:
            # Try to read the DICOM file without loading pixel data
            ds = pydicom.dcmread(file_path, stop_before_pixels=True)

            # Check if the DICOM file's Modality matches the desired modality
            if ds.Modality == modality_dicom:
                if hasattr(ds, 'ImageType') and ('LOCALIZER' in ds.ImageType or any('MIP' in entry for entry in ds.ImageType)):
                    continue
                if hasattr(ds, 'AcquisitionNumber'):
                    acquisition_numbers.append(ds.AcquisitionNumber)
                dicom_files_info.append({'file_path': file_path, 'ds': ds})

        except InvalidDicomError:
            # File is not a valid DICOM; skip it
            continue
        except Exception as e:
            # Handle any other unexpected exceptions
            warning_msg = f"An error occurred while processing file {file_path}: {str(e)}"
            warnings.warn(warning_msg, DataStructureWarning)

    if len(list(set(acquisition_numbers))) > 1:
        acquisition_number = max(set(acquisition_numbers), key=acquisition_numbers.count)

        filtered_dicom_files_info = [
            item for item in dicom_files_info
            if item['ds'].AcquisitionNumber == acquisition_number
        ]

        dicom_files_info = filtered_dicom_files_info
        warning_msg = f"The series contains multiple acquisition numbers; the most frequent one will be taken."
        warnings.warn(warning_msg, DataStructureWarning)
    if modality_dicom in ['CT', 'PT', 'MR']:
        dicom_files_info = remove_duplicate_slices(dicom_files_info)
        dicom_files_info = sort_by_geometric_position(dicom_files_info)
    return dicom_files_info


def validate_z_spacing(dicom_files):
    slice_z_origin = []
    for dcm_file in dicom_files:
        slice_z_origin.append(float(dcm_file['ds'].ImagePositionPatient[2]))
    slice_z_origin = sorted(slice_z_origin)
    slice_thickness = [abs(slice_z_origin[i] - slice_z_origin[i + 1]) for i in range(len(slice_z_origin) - 1)]
    for i in range(len(slice_thickness) - 1):
        spacing_difference = abs((slice_thickness[i] - slice_thickness[i + 1]))
        spacing_threshold = 0.1
        if spacing_difference > spacing_threshold:
            error_msg = f'Inconsistent z-spacing. Absolute deviation is {spacing_difference:.3f} which is greater than {spacing_threshold:.3f} mm.'
            raise DataStructureError(error_msg)


def calc_elapsed_time(ds, decay_constant, acquisition_time, injection_time):
    frame_reference_time = float(ds.FrameReferenceTime) / 1000
    decay_during_frame = decay_constant * ds.ActualFrameDuration / 1000
    avg_count_rate_time = (1 / decay_constant) * np.log(
        decay_during_frame / (1 - np.exp(-decay_during_frame)))

    return (acquisition_time - injection_time).total_seconds() + avg_count_rate_time - frame_reference_time


def validate_pet_dicom_tags(dicom_files):
    for dcm_file in dicom_files:
        ds = dcm_file['ds']
        image_id = dcm_file['file_path']

        try:
            pat_weight = ds[(0x0010, 0x1030)].value
            if float(pat_weight) < 1:
                warning_msg = f"For patient's {image_id} image, patient's weight tag (0071, 1022) contains weight < 1kg. Patient is excluded from the analysis."
                warnings.warn(warning_msg, DataStructureWarning)

        except (KeyError, TypeError):
            warning_msg = f"For patient's {image_id} image, patient's weight tag (0071, 1022) is not present. Patient is excluded from the analysis."
            warnings.warn(warning_msg, DataStructureWarning)
        if 'DECY' not in ds[(0x0028, 0x0051)].value or 'ATTN' not in ds[(0x0028, 0x0051)].value:
            warning_msg = f"For patient's {image_id} image, in DICOM tag (0028, 0051) either no 'DECY' (decay correction) or 'ATTN' (attenuation correction). Patient is excluded from the analysis."
            warnings.warn(warning_msg, DataStructureWarning)

        if ds.Units == 'BQML':
            injection_time = parse_time(
                ds.RadiopharmaceuticalInformationSequence[0].RadiopharmaceuticalStartTime)
            half_life = float(ds.RadiopharmaceuticalInformationSequence[0].RadionuclideHalfLife)
            decay_constant = np.log(2) / half_life

            if ds.DecayCorrection == 'START':
                if 'PHILIPS' in ds.Manufacturer.upper():

                    acquisition_time = parse_time(ds.AcquisitionTime).replace(year=injection_time.year,
                                                                              month=injection_time.month,
                                                                              day=injection_time.day)

                    elapsed_time = calc_elapsed_time(ds, decay_constant, acquisition_time, injection_time)
                elif 'SIEMENS' in ds.Manufacturer.upper() or 'CPS' in ds.Manufacturer.upper():
                    try:
                        elapsed_time = (
                                    parse_time(ds[(0x0071, 0x1022)].value).replace(year=injection_time.year,
                                                                                   month=injection_time.month,
                                                                                   day=injection_time.day)
                                    - injection_time).total_seconds()

                    except (KeyError, TypeError):
                        acquisition_time = parse_time(ds.AcquisitionTime).replace(year=injection_time.year,
                                                                                  month=injection_time.month,
                                                                                  day=injection_time.day)

                        elapsed_time = calc_elapsed_time(ds, decay_constant, acquisition_time, injection_time)
                elif 'GE' in ds.Manufacturer.upper():
                    try:
                        elapsed_time = (
                                    parse_time(ds[(0x0009, 0x100d)].value).replace(year=injection_time.year,
                                                                                   month=injection_time.month,
                                                                                   day=injection_time.day)
                                    - injection_time).total_seconds()
                    except (KeyError, TypeError):
                        acquisition_time = parse_time(ds.AcquisitionTime).replace(year=injection_time.year,
                                                                                  month=injection_time.month,
                                                                                  day=injection_time.day)
                        frame_reference_time = float(ds.FrameReferenceTime) / 1000

                        elapsed_time = (acquisition_time - injection_time).total_seconds() - frame_reference_time
                else:
                    warning_msg = f"For patient's {image_id} image, an unknown PET scaner manufacturer is present. Z-Rad only supports Philips, Siemens, and GE."
                    raise DataStructureError(warning_msg)

            elif ds.DecayCorrection == 'ADMIN':
                elapsed_time = 0
            else:
                warning_msg = f"For patient's {image_id} image, An unsupported Decay Correction {ds.DecayCorrection} is present. Only supported are 'START' and 'ADMIN'. Patient is excluded from the analysis."
                raise DataStructureError(warning_msg)
            if elapsed_time < 0:
                error_msg = f"For patient's {image_id} image, patient is excluded from the analysis due to the negative time difference in the decay factor."
                raise DataStructureError(error_msg)
            elif elapsed_time > 0 and abs(elapsed_time) < 1800 and ds.DecayCorrection != 'ADMIN':
                warning_msg = f"Only {abs(elapsed_time) / 60} minutes after the injection."
                warnings.warn(warning_msg, DataStructureWarning)
        elif ds.Units == 'CNTS' and 'PHILIPS' in ds.Manufacturer.upper():
            if not (((0x7053, 0x1009) in ds and ds[(0x7053, 0x1009)].value != 0) or ((0x7053, 0x1000) in ds and ds[(0x7053, 0x1000)].value != 0)):
                error_msg = f"For patient's {image_id} image, patient is excluded, Philips scale factors not present (PET units CNTS)"
                raise DataStructureError(error_msg)
        elif ds.Units == 'GML':
            if (0x0054, 0x1006) in ds:
                if ds[0x0054, 0x1006].value != 'BW':
                    error_msg = f"For patient's {image_id} image, patient is excluded, SUV Type is not BW (GML units)"
                    raise DataStructureError(error_msg)
        else:
            error_msg = f"For patient's {image_id} image, patient is excluded, only supported PET Units are BQML for Philips, Siemens and GE or CNTS for Philips"
            raise DataStructureError(error_msg)


def apply_suv_correction(dicom_files, suv_image):

    def process_single_slice(dicom_file_path):

        def process_gml(pixel_array_units):
            return pixel_array_units

        def process_bqml(activity_concentration, ds):
            injection_time = parse_time(ds.RadiopharmaceuticalInformationSequence[0].RadiopharmaceuticalStartTime)
            patient_weight = float(ds.PatientWeight)
            injected_dose = float(ds.RadiopharmaceuticalInformationSequence[0].RadionuclideTotalDose)
            if injected_dose <= 0:
                error_msg = f"The injected PET tracer dose is zero."
                raise DataStructureError(error_msg)
            half_life = float(ds.RadiopharmaceuticalInformationSequence[0].RadionuclideHalfLife)
            decay_constant = np.log(2) / half_life
            manufacturer = ds.Manufacturer.upper()

            if ds.DecayCorrection == 'START':
                if 'PHILIPS' in manufacturer:
                    acquisition_time = parse_time(ds.AcquisitionTime).replace(year=injection_time.year,
                                                                              month=injection_time.month,
                                                                              day=injection_time.day)

                    elapsed_time = calc_elapsed_time(ds, decay_constant, acquisition_time, injection_time)

                elif 'SIEMENS' in manufacturer or 'CPS' in manufacturer:
                    try:
                        elapsed_time = (
                                parse_time(ds[(0x0071, 0x1022)].value).replace(year=injection_time.year,
                                                                               month=injection_time.month,
                                                                               day=injection_time.day)
                                - injection_time).total_seconds()

                    except (KeyError, TypeError):
                        acquisition_time = parse_time(ds.AcquisitionTime).replace(year=injection_time.year,
                                                                                  month=injection_time.month,
                                                                                  day=injection_time.day)

                        elapsed_time = calc_elapsed_time(ds, decay_constant, acquisition_time, injection_time)
                elif 'GE' in manufacturer:
                    try:
                        elapsed_time = (
                                parse_time(ds[(0x0009, 0x100d)].value).replace(year=injection_time.year,
                                                                               month=injection_time.month,
                                                                               day=injection_time.day)
                                - injection_time).total_seconds()
                    except (KeyError, TypeError):
                        acquisition_time = parse_time(ds.AcquisitionTime).replace(year=injection_time.year,
                                                                                  month=injection_time.month,
                                                                                  day=injection_time.day)
                        frame_reference_time = float(ds.FrameReferenceTime) / 1000

                        elapsed_time = (acquisition_time - injection_time).total_seconds() - frame_reference_time
                else:
                    error_msg = f"Vendor {ds.Manufacturer} is not supported with BQML units!"
                    raise DataStructureError(error_msg)
            elif ds.DecayCorrection == 'ADMIN':
                elapsed_time = 0
            else:
                error_msg = f"Decay correction {ds.DecayCorrection} is not supported!"
                raise DataStructureError(error_msg)

            decay_factor = np.exp(-1 * ((np.log(2) * elapsed_time) / half_life))
            decay_corrected_dose = injected_dose * decay_factor
            suv = activity_concentration / (decay_corrected_dose / (patient_weight * 1000))

            return suv

        def process_cnts(pixel_array_units, ds):
            if 'PHILIPS' in ds.Manufacturer.upper():

                if (0x7053, 0x1009) in ds and ds[(0x7053, 0x1009)].value != 0:
                    activity_concentration_bqml = pixel_array_units * ds[(0x7053, 0x1009)].value
                    suv = process_bqml(activity_concentration_bqml, ds)

                    return suv

                elif (0x7053, 0x1000) in ds and ds[(0x7053, 0x1000)].value != 0:
                    suv = pixel_array_units * ds[(0x7053, 0x1000)].value

                    return suv

                else:
                    error_msg = f"Philips-specific scaling factors not present!"
                    raise DataStructureError(error_msg)

            else:
                error_msg = f"Vendor {ds.Manufacturer} is not supported with CNTS units!"
                raise DataStructureError(error_msg)

        ds = pydicom.dcmread(dicom_file_path)
        units = ds.Units
        pixel_array_units = (ds.pixel_array * ds.RescaleSlope) + ds.RescaleIntercept

        if units == 'GML':
            if (0x0054, 0x1006) in ds:
                if ds[0x0054, 0x1006].value == 'BW':
                    suv = process_gml(pixel_array_units)
                else:
                    error_msg = f"GML with {ds[0x0054, 0x1006].value} SUV normalizatoin is not supported!"
                    raise DataStructureError(error_msg)
            else:
                suv = process_gml(pixel_array_units)
        elif units == 'BQML':
            suv = process_bqml(pixel_array_units, ds)
        elif units == 'CNTS':
            suv = process_cnts(pixel_array_units, ds)
        else:
            error_msg = f"Units {units} are not supported!"
            raise DataStructureError(error_msg)

        return suv.T

    intensity_array = np.zeros(suv_image.GetSize())

    dicom_files = sorted(dicom_files, key=lambda f: float(f['ds'].ImagePositionPatient[2]))
    for z_slice_id, dicom_file in enumerate(dicom_files):
        intensity_array[:, :, z_slice_id] = process_single_slice(dicom_file['file_path'])

    # Flip from head to toe
    intensity_image = sitk.GetImageFromArray(intensity_array.T)
    intensity_image.SetOrigin(suv_image.GetOrigin())
    intensity_image.SetSpacing(np.array(suv_image.GetSpacing()))
    intensity_image.SetDirection(np.array(suv_image.GetDirection()))
    return intensity_image

def modality_mapping(modality):
    modality_map = {'PET': 'PT', 'CT': 'CT', 'MRI': 'MR', 'RTSTRUCT': 'RTSTRUCT', 'MG': 'MG', 'RTDOSE': 'RTDOSE'}
    return modality_map[modality]

def process_dicom_series(dicom_files):
    reader = sitk.ImageSeriesReader()
    file_names = [i['file_path'] for i in dicom_files]
    reader.SetFileNames(file_names)
    image = reader.Execute()

    slice_z_origin = []
    direction = None
    for dicom_file in dicom_files:
        ds = dicom_file['ds']

        if ds.Modality in ['CT', 'PT', 'MR']:
            pixel_spacing = ds.PixelSpacing

            # Use full 3D position
            iop = np.array(ds.ImageOrientationPatient, dtype=float)  # 6 values
            row_cosines = iop[:3]
            col_cosines = iop[3:]
            normal = np.cross(row_cosines, col_cosines)  # image plane normal
            if direction is None:
                direction = np.vstack([row_cosines, col_cosines, normal]).flatten(order="F")
            # Project position onto normal vector
            distance_along_normal = np.dot(np.array(ds.ImagePositionPatient, dtype=float), normal)
            slice_z_origin.append(distance_along_normal)

        elif ds.Modality == 'MG':
            pixel_spacing = ds.ImagerPixelSpacing
            slice_z_origin.append(ds.BodyPartThickness)
            image.SetOrigin([0, 0, 0])
            direction = [1, 0, 0, 0, 1, 0, 0, 0, 1]

    slice_z_origin = sorted(slice_z_origin)
    if len(slice_z_origin) > 1:
        slice_thickness = np.median(np.abs(np.diff(np.asarray(slice_z_origin, float))))

    elif len(slice_z_origin) == 1:
        slice_thickness = slice_z_origin[0]

    image.SetSpacing((float(pixel_spacing[0]), float(pixel_spacing[1]), float(slice_thickness)))
    image.SetDirection(direction)
    if dicom_files[0]['ds'].Modality == 'CT' and np.min(sitk.GetArrayFromImage(image)) >= 0:
        error_msg = f'Non-negative CT intensity. SITK failed to convert CT into HU for {dicom_files[0]["file_path"]}. The patient is excluded from analysis'
        raise DataStructureError(error_msg)

    return image


def extract_dicom_mask(rtstruct_path, roi_name, image):
    def generate_mask_array(contours, sitk_image):

        # Get image dimensions: (width, height, depth)
        width, height, depth = sitk_image.GetSize()
        # Create an empty 3D mask with shape (depth, height, width)
        mask_array = np.zeros((depth, height, width), dtype=np.uint8)

        for contour in contours:
            # Normalize contour type string
            contour_type = contour['type'].upper().replace('_', '').strip()
            if contour_type not in ['CLOSEDPLANAR', 'INTERPOLATEDPLANAR', 'CLOSEDPLANARXOR']:
                continue

            points = contour['points']
            num_points = len(points['x'])
            # Transform all physical points to continuous index coordinates
            transformed_points = np.array([
                sitk_image.TransformPhysicalPointToContinuousIndex(
                    (points['x'][i], points['y'][i], points['z'][i])
                )
                for i in range(num_points)
            ])

            # Compute rounded z indices for each point
            z_indices = np.rint(transformed_points[:, 2]).astype(int)

            # Prepare polygon coordinates for x and y. Note: polygon2mask expects (row, col) = (y, x)
            polygon_coords = np.column_stack((transformed_points[:, 1], transformed_points[:, 0]))
            mask_layer = draw.polygon2mask((height, width), polygon_coords)

            # Combine the mask on each relevant slice:
            # Use XOR for "CLOSEDPLANARXOR" contours; for others, simply add (logical OR).
            for z in z_indices:
                if contour_type == 'CLOSEDPLANARXOR':
                    mask_array[z] = np.logical_xor(mask_array[z], mask_layer).astype(np.uint8)
                else:
                    mask_array[z] = np.logical_or(mask_array[z], mask_layer).astype(np.uint8)

        return mask_array

    def get_contour_data(file_path, selected_roi):
        def get_contour_coord(metadata, current_roi_sequence, skip_contours_bool=False):
            contour_info = {
                'name': getattr(metadata[current_roi_sequence.ReferencedROINumber], 'ROIName', 'unknown'),
                'roi_number': current_roi_sequence.ReferencedROINumber,
                'referenced_frame': getattr(metadata[current_roi_sequence.ReferencedROINumber],
                                            'ReferencedFrameOfReferenceUID', 'unknown'),
                'display_color': getattr(current_roi_sequence, 'ROIDisplayColor', [])
            }

            if not skip_contours_bool and hasattr(current_roi_sequence, 'ContourSequence'):
                contour_info['sequence'] = [{
                    'type': getattr(contour, 'ContourGeometricType', 'unknown'),
                    'points': {
                        'x': [contour.ContourData[i] for i in range(0, len(contour.ContourData), 3)],
                        'y': [contour.ContourData[i + 1] for i in range(0, len(contour.ContourData), 3)],
                        'z': [contour.ContourData[i + 2] for i in range(0, len(contour.ContourData), 3)]
                    }
                } for contour in current_roi_sequence.ContourSequence]

            return contour_info

        dicom_data = pydicom.dcmread(file_path)
        if not hasattr(dicom_data, 'StructureSetROISequence'):
            raise InvalidDicomError()

        contour_data = None
        metadata_map = {data.ROINumber: data for data in dicom_data.StructureSetROISequence}
        for roi_sequence in dicom_data.ROIContourSequence:
            roi_name = getattr(metadata_map[roi_sequence.ReferencedROINumber], 'ROIName', 'unknown')
            if roi_name == selected_roi:
                contour_data = get_contour_coord(metadata_map, roi_sequence)
                break

        return contour_data

    rt_struct = get_contour_data(rtstruct_path, roi_name)
    if rt_struct:
        mask = generate_mask_array(rt_struct['sequence'], image)
        return Image(array=mask,
                     origin=image.GetOrigin(),
                     spacing=np.array(image.GetSpacing()),
                     direction=image.GetDirection(),
                     shape=image.GetSize())
    else:
        return Image()


def get_all_structure_names(rtstruct_path):
    """Extracts all structure names from an RTSTRUCT DICOM file.

    Args:
        rtstruct_path (str): Path to the RTSTRUCT DICOM file.

    Returns:
        list: A list of structure names.

    Raises:
        InvalidDicomError: If the DICOM file does not have a StructureSetROISequence attribute.
    """
    # Read the DICOM file
    dicom_data = pydicom.dcmread(rtstruct_path)

    # Check if the file contains a StructureSetROISequence
    if not hasattr(dicom_data, 'StructureSetROISequence'):
        raise InvalidDicomError(f"The DICOM file at {rtstruct_path} is not a valid RTSTRUCT file.")

    # Map ROI numbers to metadata for quick lookup
    metadata_map = {roi_data.ROINumber: roi_data for roi_data in dicom_data.StructureSetROISequence}

    # Extract structure names
    structure_names = []
    if hasattr(dicom_data, 'ROIContourSequence'):
        for roi_sequence in dicom_data.ROIContourSequence:
            roi_number = roi_sequence.ReferencedROINumber
            roi_name = getattr(metadata_map.get(roi_number, {}), 'ROIName', 'unknown')
            structure_names.append(roi_name)

    return structure_names
